import json
import os
import re
import sys
from pathlib import Path
from argparse import (ArgumentParser, ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter)
from datetime import datetime
from typing import Optional
from dataclasses import dataclass
import pandas as pd
from functools import partial
import logging
import errno
from locidex.classes.blast import BlastSearch, FilterOptions
from locidex.classes.db import search_db_conf, db_config
from locidex.manifest import DBData
from locidex.classes.seq_intake import seq_intake, seq_store, HitFilters
from locidex.constants import BlastCommands, FILE_TYPES, BlastColumns, DB_EXPECTED_FILES, OPTION_GROUPS, DBConfig
from locidex.utils import write_seq_dict, check_db_groups, slots, get_format
from locidex.version import __version__

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.DEBUG)

def add_args(parser=None):
    if parser is None:
        parser = ArgumentParser(
            description="Locidex: Advanced searching and filtering of sequence databases using query sequences",)
    parser.add_argument('-q','--query', type=str, required=True,help='Query sequence file')
    parser.add_argument('-o', '--outdir', type=str, required=True, help='Output directory to put results')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-d', '--db', type=str, required=False, help='Locidex database directory')
    group.add_argument("--db_group", type=str, required=False, help="A directory of databases containing a manifest file. Requires the db_name option to be set to select the correct db")
    parser.add_argument('-n', '--name', type=str, required=False, help='Sample name to include default=filename')
    parser.add_argument('-c', '--config', type=str, required=False, help='Locidex parameter config file (json)')
    parser.add_argument('--db_name', type=str, required=False, help='Name of database to perform search, used when a manifest is specified as a db')
    parser.add_argument('--db_version', type=str, required=False, help='Version of database to perform search, used when a manifest is specified as a db')
    parser.add_argument('--min_evalue', type=float, required=False, help='Minumum evalue required for match',
                        default=0.0001)
    parser.add_argument('--min_dna_len', type=int, required=False, help='Global minumum query length dna',
                        default=1)
    parser.add_argument('--min_aa_len', type=int, required=False, help='Global minumum query length aa',
                        default=1)
    parser.add_argument('--max_dna_len', type=int, required=False, help='Global maximum query length dna',
                        default=10000000)
    parser.add_argument('--max_aa_len', type=int, required=False, help='Global maximum query length aa',
                        default=10000000)
    parser.add_argument('--min_dna_ident', type=float, required=False, help='Global minumum DNA percent identity required for match',
                        default=80.0)
    parser.add_argument('--min_aa_ident', type=float, required=False, help='Global minumum AA percent identity required for match',
                        default=80.0)
    parser.add_argument('--min_dna_match_cov', type=float, required=False, help='Global minumum DNA percent hit coverage identity required for match',
                        default=80.0)
    parser.add_argument('--min_aa_match_cov', type=float, required=False, help='Global minumum AA percent hit coverage identity required for match',
                        default=80.0)
    parser.add_argument('--max_target_seqs', type=int, required=False, help='Maximum number of hit seqs per query',
                        default=10)
    parser.add_argument('--override', required=False, help='Overwrite individual loci thresholds for filtering and use the global parameters',
                        action='store_true')
    parser.add_argument('--n_threads','-t', type=int, required=False,
                        help='CPU Threads to use', default=1)
    parser.add_argument('--format', type=str, required=False,
                        help='Format of query file [genbank,fasta]')
    parser.add_argument('--translation_table', type=int, required=False,
                        help='Table to use for translation', default=11)
    parser.add_argument('-a', '--annotate', required=False, help='Perform annotation on unannotated input fasta',
                        action='store_true')
    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)
    parser.add_argument('-f', '--force', required=False, help='Overwrite existing directory',
                        action='store_true')
    return parser


def create_fasta_from_df(df,label_col,seq_col,out_file):
    return write_seq_dict(dict(zip(df[label_col].tolist(), df[seq_col])), out_file)

@dataclass
class DefaultSearchOpts:
    program: str
    seq_col: str
    pident_filter: FilterOptions
    qcovs_filter: FilterOptions
    db_dir: Path
    output_dir: str


def create_outputs(output_dir: Path, db_data: DBData, blast_params: dict, configuration: DefaultSearchOpts, filtered_df: pd.DataFrame, filter_options: dict) -> pd.DataFrame:
    """
    Create outputs of blast hits
    output_dir Path: output location of search data
    configuration DefaultSearchOpts: Pararmeters passed to 'run_search' from the cli

    This function will have some needed clean up once the cli is tidied
    """
    hsps_out = "hsps.txt" #? Need to follow up on what hsps stands for, hsps is a blast that needs to included
    label_col = 'index'
    output_directory = output_dir.joinpath(configuration.output_dir)
    query_fasta = output_directory.joinpath("queries.fasta")
    output_hsps = output_directory.joinpath(hsps_out)
    if not output_directory.exists() or not output_directory.is_dir():
        os.makedirs(output_directory, 0o755, exist_ok=True) # exist okay is true, as the program will error out before it clobbers the directory

    output_file = create_fasta_from_df(filtered_df, label_col=label_col, seq_col=configuration.seq_col, out_file=query_fasta)
    #! Blast search output columns should be showing up
    search_data = BlastSearch(db_data, output_file, blast_params, configuration.program, BlastColumns._fields, filter_options)
    searched_df = search_data.get_blast_data(configuration.db_dir, output_hsps)
    return searched_df

def run_search(config):

    # Input Parameters
    query_file = Path(config['query'])
    outdir = Path(config['outdir'])
    db_dir = config['db']
    min_dna_ident = config['min_dna_ident']
    min_aa_ident =config['min_aa_ident']
    min_evalue = config['min_evalue']
    min_dna_match_cov = config['min_dna_match_cov']
    min_aa_match_cov = config['min_aa_match_cov']
    n_threads = config['n_threads']
    translation_table = config['translation_table']
    force = config['force']
    format = config['format']
    min_dna_len = config['min_dna_len']
    max_dna_len = config['max_dna_len']
    min_aa_len = config['min_aa_len']
    max_aa_len = config['max_aa_len']
    sample_name = config['name']
    perform_annotation = config['annotate']
    max_target_seqs = config['max_target_seqs']
    override = config['override']

    if max_count := config.get('max_ambig_count'):
        max_ambig_count = max_count
    else:
        max_ambig_count = float('inf')

    if not perform_annotation:
        perform_annotation = False

    if sample_name == None:
        sample_name = query_file.stem

    db_data = DBData(db_dir=db_dir)
    config = config | db_data.config_data.to_dict()
    run_data = dict()
    run_data['analysis_start_time'] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    run_data['parameters'] = config

    if os.path.isdir(outdir) and not force:
        logger.critical(f'Error {outdir} exists, if you would like to overwrite, then specify --force')
        raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST), str(outdir))

    if not os.path.isdir(outdir):
        os.makedirs(outdir, 0o755)

    if format is None:
        format = get_format(query_file)
    else:
        format = format.lower()

    if format is None or format not in FILE_TYPES:
        if format is None:
            logger.critical(f'Could not guess format for {query_file}')
            raise ValueError("Could not determine input type for: {}".format(query_file))
        else:
            logger.critical(f'Format for query file must be one of {list(FILE_TYPES.keys())}, you supplied {format}')
            raise ValueError("Format supplied for {} is incorrect.".format(format))

    seq_obj = seq_intake(input_file=query_file, file_type=format, feat_key='CDS', 
                        translation_table=translation_table, perform_annotation=perform_annotation)

    if perform_annotation:
        gbk_data = []
        for idx,genes in enumerate(seq_obj.prodigal_genes):
            f = os.path.join(outdir,f"annotations_{idx}.gbk")
            with open(f,'w') as oh:
                genes.write_genbank(oh, sequence_id=f'{sample_name}_{idx}')
            gbk_data.append(Path(f).read_text())
            os.remove(f)
        f = os.path.join(outdir, f"annotations.gbk")
        with open(f, 'w') as oh:
            oh.write("\n".join([str(x) for x in gbk_data]))

    blast_params = {
        'evalue': min_evalue,
        'max_target_seqs': max_target_seqs,
        'num_threads': n_threads,
    }

    filter_options = {
        'evalue': FilterOptions(min=None, max=min_evalue, include=None)
    }


    df = pd.DataFrame.from_dict(seq_obj.seq_data)
    filtered_df = df
    filtered_df['index'] = filtered_df.index.to_list()
    hit_filters = HitFilters(
        min_dna_len = min_dna_len,
        max_dna_len=max_dna_len,
        min_dna_ident=min_dna_ident,
        min_dna_match_cov=min_dna_match_cov,
        min_aa_len=min_aa_len,
        max_aa_len=max_aa_len,
        min_aa_ident=min_aa_ident,
        min_aa_match_cov=min_aa_match_cov,
        dna_ambig_count=max_ambig_count)
    

    store_obj = seq_store(sample_name, db_data.config_data, db_data.metadata['meta'],
                        seq_obj.seq_data, BlastColumns._fields, hit_filters,override=override)

    protein_filter = DefaultSearchOpts(
        program=BlastCommands.blastp, 
        seq_col="aa_seq", 
        pident_filter=FilterOptions(min=min_aa_ident, max=None, include=None),
        qcovs_filter=FilterOptions(min=min_aa_match_cov, max=None, include=None),
        db_dir=db_data.protein_blast_db,
        output_dir=DBData.protein_name())
    
    nucleotide_filter = DefaultSearchOpts(
        program=BlastCommands.blastn, 
        seq_col="dna_seq", 
        pident_filter=FilterOptions(min=min_dna_ident, max=None, include=None),
        qcovs_filter=FilterOptions(min=min_dna_match_cov, max=None, include=None),
        db_dir=db_data.nucleotide_blast_db,
        output_dir=DBData.nucleotide_name())

    searched_hits_col = 'qseqid'
    output_creation = partial(create_outputs, output_dir=outdir, db_data=db_data, filtered_df=filtered_df, blast_params=blast_params, filter_options=filter_options)
    if db_data.nucleotide:
        searched_data = output_creation(configuration=nucleotide_filter)
        store_obj.add_hit_data(searched_data, DBData.nucleotide_name(), searched_hits_col)
    if db_data.protein:
        searched_data = output_creation(configuration=protein_filter)
        store_obj.add_hit_data(searched_data, DBData.protein_name(), searched_hits_col)

    store_obj.filter_hits()
    store_obj.convert_profile_to_list()
    run_data['result_file'] = str(outdir.joinpath("seq_store.json"))

    with open(run_data['result_file'], "w") as fh:
        fh.write(json.dumps(store_obj.record, indent=4))

    run_data['analysis_end_time'] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

    with open(outdir.joinpath("run.json"),'w' ) as fh:
        fh.write(json.dumps(run_data, indent=4))


def run(cmd_args=None):
    if cmd_args is None:
        parser = add_args()
        cmd_args = parser.parse_args()
    analysis_parameters = vars(cmd_args)

    analysis_parameters = check_db_groups(analysis_params=analysis_parameters, cmd_args=cmd_args)
    
    config_file = cmd_args.config

    config = {}
    if config_file is not None:
        with open(config_file) as fh:
            config = json.loads(fh.read())

    for p in analysis_parameters:
        if not p in config:
            config[p] = analysis_parameters[p]

    run_search(config)


# call main function
if __name__ == '__main__':
    run()






import pytest
import errno
import json
import csv
from locidex import merge
from os import listdir



DUPLICATE_NAMES = [
    "locidex/example/merge/merge_in/report.json",
    "locidex/example/merge/merge_in/report2.json",
]

MERGE_SUCCESSFULLY = [
    "locidex/example/merge/merge_in/report.json",
    "locidex/example/merge/merge_in/report1.json",
]

MERGE_MISMATCH_PROFILES = "locidex/example/merge/merge_inputassure/sample1.mlst.json"


def test_throws_duplicate_error():
    """
    """
    with pytest.raises(ValueError, match="Attempting to merge allele profiles with the same sample name: 1"):
        merge.read_file_list(DUPLICATE_NAMES, perform_db_validation=True)

def test_merge_profiles_no_error():
    merge.read_file_list(MERGE_SUCCESSFULLY, perform_db_validation=True)


def test_check_files_exist():
    with pytest.raises(FileNotFoundError, match=str(errno.ENOENT)):
        fail = "dne.txt"
        merge.check_files_exist([fail])

    merge.check_files_exist(MERGE_SUCCESSFULLY)

def test_extract_profiles():
    records, compare_error = merge.read_file_list(MERGE_SUCCESSFULLY, perform_db_validation=True)
    extracted_profiles = merge.extract_profiles(records)
    assert len(extracted_profiles) == 2
    value1, value2 = extracted_profiles.values()
    assert value1 == value2
    assert value1 == {'aroC': '9048803cd72dee3c868cd2dc5dc5650d', 'dnaN': '2772ad8b8e0f7b50f1396c31fbe53f2d', 'hemD': '620f99723c4e190abe096b11ca34b944', 'hisD': '38027ac1ac34817584a176c7e575e97e', 'purE': '9855cbf4009439498bf84cacefce4d8f', 'sucA': '9289fc07cc8e93cfe0716e6f613cefdb', 'thrA': '9e1aa76bb42279ed7ec8fc30f984b65d'}

def test_profile_validation():
    mlst_file = json.load(open(MERGE_MISMATCH_PROFILES))
    ## Test that if the sample name matches the profile in the MLST json file nothing is changed
    new_mlst, mlst_report = merge.validate_and_fix_profiles(mlst_file, "sampleA", "sample1.mlst.json")
    assert new_mlst["data"]["profile"] == {'sampleA': {'l1': '1', 'l2': '1', 'l3': '1'}}
    assert mlst_report == None
    ## Test that a different sample than the profile in the MLST json will be changed
    same_mlst, mlst_report = merge.validate_and_fix_profiles(mlst_file, "sample1", "sample1.mlst.json")
    assert same_mlst["data"]["profile"] == {'sample1': {'l1': '1', 'l2': '1', 'l3': '1'}}
    assert mlst_report[2] == "sample1 ID and JSON key in sample1.mlst.json DO NOT MATCH. The 'sampleA' key in sample1.mlst.json has been forcefully changed to 'sample1': User should manually check input files to ensure correctness."

def test_profile_validation_report(tmpdir):
    """ Note: This test is an end-to-end test for locidex merge by default as it looks for the final outputs.
    The input however was created to check the MLST_error_report.csv file which reports the renaming. In theory other end-to-end checks can be added here.
    """
    CONFIG = {'command': 'merge',
                                'input': [[
                                    'locidex/example/merge/merge_inputassure/sampleQ.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample1.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample2.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample3.mlst.json']],
                                'outdir': f"{tmpdir}",
                                'strict': False,
                                'force': True,
                                'profile_ref': 'locidex/example/merge/merge_inputassure/test_ref_profile.csv',
                                'loci': None}
    merge.run_merge(CONFIG)
    merge_output = sorted(listdir(tmpdir))
    assert merge_output == ["MLST_error_report.csv", "profile.tsv", "run.json"]
    with open(f"{tmpdir}/MLST_error_report.csv", "r", newline='') as file:
        reader_obj = csv.reader(file)
        next(reader_obj)  # Skip the first line (header)
        second_line = next(reader_obj)
        assert second_line == ['sample1',"['sampleA']", "sample1 ID and JSON key in sample1.mlst.json DO NOT MATCH. The 'sampleA' key in sample1.mlst.json has been forcefully changed to 'sample1': User should manually check input files to ensure correctness."]
    with open(f"{tmpdir}/profile.tsv", "r", newline='') as file:
        reader_obj = csv.reader(file, delimiter="\t",)
        next(reader_obj)  # Skip the first line (header)
        sampleQ = next(reader_obj)
        assert sampleQ == ['sampleQ','1','2','1']
        sample1 = next(reader_obj)
        assert sample1 == ['sample1','1','1','1']

def test_profile_validation_noreport(tmpdir):
    """ Note: Similar to test_profile_validation_report, except checking that functionality stays the same without a report
    """
    CONFIG = {'command': 'merge',
                                'input': [[
                                    'locidex/example/merge/merge_inputassure/sampleQ.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample1.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample2.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample3.mlst.json']],
                                'outdir': f"{tmpdir}",
                                'strict': False,
                                'force': True,
                                'profile_ref': None,
                                'loci': None}
    merge.run_merge(CONFIG)
    merge_output = sorted(listdir(tmpdir))
    assert merge_output == ["profile.tsv", "run.json"]

    with open(f"{tmpdir}/profile.tsv", "r", newline='') as file:
        reader_obj = csv.reader(file, delimiter="\t",)
        next(reader_obj)  # Skip the first line (header)
        sampleQ = next(reader_obj)
        assert sampleQ == ['sampleQ','1','2','1']
        sample1 = next(reader_obj)
        assert sample1 == ['sampleA','1','1','1']

def test_removal_loci():
    """
    Test that the --loci maintains the loci specified to subset the profiles properly.
    Both to remove 'aroC': '9048803cd72dee3c868cd2dc5dc5650d' and to ensure the order for the remaining loci follows the loci_to_keep.
    """
    records, compare_error = merge.read_file_list(MERGE_SUCCESSFULLY, perform_db_validation=True, loci_to_keep=[ 'sucA',  'purE', 'dnaN', 'hemD', 'hisD', 'thrA'])
    extracted_profiles = merge.extract_profiles(records)
    assert len(extracted_profiles) == 2
    value1, value2 = extracted_profiles.values()
    assert value1 == value2
    assert value1 == {'sucA': '9289fc07cc8e93cfe0716e6f613cefdb', 'purE': '9855cbf4009439498bf84cacefce4d8f', 'dnaN': '2772ad8b8e0f7b50f1396c31fbe53f2d', 'hemD': '620f99723c4e190abe096b11ca34b944', 'hisD': '38027ac1ac34817584a176c7e575e97e', 'thrA': '9e1aa76bb42279ed7ec8fc30f984b65d'}

def test_removal_loci_argument(capfd, tmpdir):
    """ Note: test that the --loci argument works to take in a file with loci to keep
    """
    CONFIG = {'command': 'merge',
                                'input': [[
                                    'locidex/example/merge/merge_inputassure/sampleQ.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample1.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample2.mlst.json',
                                    'locidex/example/merge/merge_inputassure/sample3.mlst.json']],
                                'outdir': f"{tmpdir}",
                                'strict': False,
                                'force': True,
                                'profile_ref': None,
                                'loci': 'locidex/example/merge/merge_inputassure/loci_to_keep.txt'}
    merge.run_merge(CONFIG)
    out, err = capfd.readouterr()
    assert out == "INFO:locidex merge: Keeping only the following loci: ['l1', 'aroC', 'dnaN']\n"


""""
Tests for locidex build

"""

import pytest, warnings
from locidex import build

import os
import json
from dataclasses import dataclass

TEST_DATA = "locidex/example/build_db_mlst_in/senterica.mlst.txt"
EXPECTED_DATA_OUT = "locidex/example/build_db_mlst_out"
EXCLUDE_COMP_FILES = frozenset(["protein.pjs", "nucleotide.njs", "nucleotide.nin", "protein.pin", "config.json", "results.json"])


@dataclass
class CMDArgs:
    input_file: os.PathLike
    outdir: os.PathLike
    name: str
    db_ver: str
    db_desc: str
    author: str
    date: str
    force: bool

@pytest.fixture(scope="module")
def output_directory(tmp_path_factory):
    fn = tmp_path_factory.mktemp("build")
    return str(fn)

@pytest.fixture(scope="module")
def cmd_args(output_directory):
    command = CMDArgs(input_file=TEST_DATA, 
                    outdir=output_directory,
                    name='Locidex Database',
                    db_ver='1.0.0',
                    db_desc='test',
                    force=True,
                    author='mw',
                    date=''
                    )
    return command

def test_build(cmd_args):
    """
    Test that no errors are raised
    """
    build.run(cmd_args)
    assert len([file  for file in os.listdir(cmd_args.outdir) if "json" in file]) > 0

def test_build_outputs(output_directory):
    """
    Test initial directory structure is the same
    """
    output_data = os.listdir(output_directory)
    expected_data = os.listdir(EXPECTED_DATA_OUT)
    assert sorted(output_data) == sorted(expected_data)

def get_all_file_paths(dir):
    """
    List all files relative paths in a directory
    """
    paths = []
    for i in os.walk(dir):
        paths.extend([(i[0], g) for g in i[2]])
    return paths

@pytest.mark.parametrize("f_name,comp_fields,primary_key", [
    ("config.json", ["db_name", "db_version", "db_author", "db_desc", 
                "db_num_seqs", "is_nucl", "is_prot", "nucleotide_db_name",
                "protein_db_name"], None),
    ("results.json", ["input_file", "name", "db_ver", "db_desc", "author", "force"], "parameters")
])
def test_config_results_json(output_directory,f_name,comp_fields, primary_key):
    """Verify that config and results files outputs are the same.

    Can probably make this exclusionary fields instead of inclusion. But I want to check that fields provided still exist.
    """
    config_file_name = f_name
    actual = os.path.join(output_directory, config_file_name)
    expected = os.path.join(EXPECTED_DATA_OUT, config_file_name)
    with open(actual, 'r', encoding='utf8') as act, open(expected, 'r', encoding='utf8') as expc:
        
        act_json = json.load(act)
        expc_json = json.load(expc)

        if primary_key is not None:
            act_json = act_json[primary_key]
            expc_json = expc_json[primary_key]

        for i in comp_fields:
            assert act_json.get(i) is not None and expc_json.get(i) is not None
            assert act_json[i] == expc_json[i]


def test_validate_field_continuity(output_directory):
    """Verify that like fields between the output files are the same. The expected should match if the actual does
    otherwise this would be a time to update tests.
    """

    config_files = os.path.join(output_directory, "config.json")
    results_files = os.path.join(output_directory, "results.json")

    with open(config_files, 'r', encoding='utf8') as conf, open(results_files, 'r', encoding="utf8") as results:
        results_fields = json.load(results)["parameters"]
        conf_fields = json.load(conf)
        assert results_fields["db_ver"] == conf_fields["db_version"]
        assert results_fields["name"] == conf_fields["db_name"]
        assert results_fields["author"] == conf_fields["db_author"]
        assert results_fields["db_desc"] == conf_fields["db_desc"]
        if results_fields["date"] == '':
            warnings.warn("In results.json the date field (database) is empty")
        else:
            assert results_fields["date"] == conf_fields["db_date"]


    
@pytest.mark.skip(reason="BLAST versions provide different database structure and output results. Need to fix version to 2.15.0 and try to find better testing logic")
def test_file_outputs(output_directory):
    """
    Make sure all required blast outputs are created
    Fails as blast version needs to be set
    """
    #test_build(cmd_args)
    output_data = get_all_file_paths(output_directory)
    expected_data = get_all_file_paths(EXPECTED_DATA_OUT)
    output_files = [i[1] for i in output_data]
    expected_files = [i[1] for i in expected_data]
    assert set(output_files) == set(expected_files)

    outputs_sorted = sorted(output_data, key=lambda x: x[1])
    expected_outputs_sorted = sorted(expected_data, key=lambda x: x[1])
    assert len(outputs_sorted) == len(expected_outputs_sorted)

    for test, expected in zip(outputs_sorted, expected_outputs_sorted):
        if test[1] in EXCLUDE_COMP_FILES and expected[1] in EXCLUDE_COMP_FILES:
            continue
        test_path = os.path.join(test[0], test[1])
        expected_path = os.path.join(expected[0], expected[1])
        with open(test_path, 'rb') as t_in, open(expected_path, 'rb') as e_in:
            test_text = hash(t_in.read())
            expected_text = hash(e_in.read())
            assert test_text == expected_text


import gzip
import json
import os
import re
import sys
import errno
import typing
from argparse import (ArgumentParser, ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter)
from datetime import datetime
from functools import partial
from mimetypes import guess_type
from multiprocessing import Pool, cpu_count
import logging
import pandas as pd
from locidex.classes.aligner import align, parse_align
from locidex.constants import DBConfig, raise_file_not_found_e
from locidex.report import ReportData, Data, Parameters
from locidex.version import __version__

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.INFO)

class OutputRow(typing.NamedTuple):
    sample_id: str
    keys: list[str]
    message: str

def add_args(parser=None):
    """
    TODO disabling alignment until test data is prepared
    """
    if parser is None:
        parser = ArgumentParser(
            description="Locidex merge: Concatonate set of input profile.json files into  a tsv table or aligned fasta")
    parser.add_argument('-i','--input', type=str, required=True,help='Input file to report', action='append', nargs='+')
    parser.add_argument('-o', '--outdir', type=str, required=True, help='Output file to put results')
    #parser.add_argument('--n_threads','-t', type=int, required=False,
    #                    help='CPU Threads to use', default=1)
    #parser.add_argument('--linker','-l', type=str, required=False,
    #                    help='Linker sequence for alignment', default='NNNNNNNNNNNNNNNNNNNN')
    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)
    parser.add_argument('-s', '--strict', required=False, help='Only merge data produces by the same db',
                        action='store_true')
    #parser.add_argument('-a', '--align', required=False, help='Perform alignment with individual loci to produce a concatenated alignment',
    #                    action='store_true')
    parser.add_argument('-f', '--force', required=False, help='Overwrite existing directory',
                        action='store_true')
    parser.add_argument('-p', '--profile_ref', type=str, required=False, help='Provide a TSV file with profile references for overriding MLST profiles. Columns [sample/sample_name,mlst_alleles]')
    parser.add_argument('--loci', required=False, default=None, help='Specifies a file (or command-separated list) of loci to keep from MLST files',)
    return parser


def get_file_list(input_files):
    file_list = []
    if len(input_files) > 1:
        file_list = input_files
    else:
        if re.search(".json$", input_files[0]) or re.search(".json.gz$", input_files[0]):
            file_list = input_files
        else:
            if not os.path.isfile(input_files[0]):
                logger.critical("File {} does not exists".format(input_files[0]))
                raise_file_not_found_e(input_files[0], logger=logger)

            encoding = guess_type(input_files[0])[1]
            _open = partial(gzip.open, mode='rt') if encoding == 'gzip' else open
            with _open(input_files[0]) as f:
                for line in f:
                    line = line.rstrip()
                    if not os.path.isfile(line):
                        logger.critical("Could not find file: {}".format(line))
                        raise_file_not_found_e(line, logger)
                    file_list.append(line)
    return file_list

def validate_input_file(data_in: dict, filename: str, db_version: str, db_name: str, perform_db_validation: bool, perform_profile_validation: bool, profile_refs_dict:dict) -> tuple[ReportData, str, str, typing.Optional[OutputRow]]:
    """
    Validate input data for usage verifying db_versions, db_names and MLST profiles are the same
    """
    if perform_profile_validation:
        user_provided_profile = profile_refs_dict[filename][0]
        data_in, mlst_report = validate_and_fix_profiles(data_in, user_provided_profile, filename)
    else:
        mlst_report = None

    try:
        sq_data = ReportData.deseriealize(data_in)
    except KeyError:
        logger.critical("Missing fields in configuration required fields in in reported allele file. Fields required: {}".format(ReportData.fields()))
        raise ValueError("Missing fields in configuration required fields in in reported allele file. Fields required: {}".format(ReportData.fields()))

    else:

        if db_version is not None and sq_data.db_info.db_version != db_version and perform_db_validation:
            logger.critical("You are attempting to merge files that were created using different database versions.")
            raise ValueError("You are attempting to merge files that were created using different database versions.")

        if db_name is not None and sq_data.db_info.db_name != db_name and perform_db_validation:
            logger.critical("You are attempting to merge files that have different names.")
            raise ValueError("You are attempting to merge files that have different names. {} {}".format(sq_data.db_info.db_name, db_name))

    return sq_data, sq_data.db_info.db_version, sq_data.db_info.db_name, mlst_report

def check_files_exist(file_list: list[os.PathLike]) -> None:
    """
    Verify that all files to be analyzed exist
    """
    for file in file_list:
        if not os.path.isfile(file):
            logger.critical("Could not find file: {}".format(file))
            raise_file_not_found_e(file, logger)


def read_file_list(file_list, perform_db_validation=False, perform_profile_validation=False, loci_to_keep=None):
    records = {}
    db_version = None
    db_name = None
    check_files_exist(file_list)
    # Before we can perform profile validation of the MLST files we need to check that a proper profile key file was provided
    if perform_profile_validation:
        modified_MLST_files = [["sample", "JSON_key", "error_message"]]
        check_files_exist([perform_profile_validation])
        profile_refs_dict = read_samplesheet(perform_profile_validation)
    else:
        modified_MLST_files = None
        profile_refs_dict = None

    for f in file_list:
        encoding = guess_type(f)[1]
        filename = os.path.basename(f)
        _open = partial(gzip.open, mode='rt') if encoding == 'gzip' else open
        with _open(f) as fh:
            data = json.load(fh)
            sq_data, db_version, db_name, mlst_report = validate_input_file(data,
                                                            filename,
                                                            db_version=db_version,
                                                            db_name=db_name,
                                                            perform_db_validation=perform_db_validation,
                                                            perform_profile_validation=perform_profile_validation,
                                                            profile_refs_dict = profile_refs_dict)
            if perform_profile_validation and mlst_report is not None:
                modified_MLST_files.append(mlst_report)
            sample_name = sq_data.data.sample_name
            # Remove loci that are not in loci list if provided
            if loci_to_keep is not None:
                profile = sq_data.data.profile[sample_name]
                core_loci = {k: profile[k] for k in loci_to_keep if k in profile}
                sq_data.data.profile[sample_name] = core_loci

            if records.get(sq_data.data.sample_name) is None:
                records[sample_name] = sq_data
            else:
                logger.critical("Duplicate sample name detected: {}".format(sq_data.data.sample_name))
                raise ValueError("Attempting to merge allele profiles with the same sample name: {}".format(sq_data.data.sample_name))

    return records, modified_MLST_files

def extract_profiles(records):
    profile = {}
    for id in records:
        for sample_name in records[id].data.profile:
            if profile.get(sample_name) is not None:
                logger.critical("Sample {} already exists and will not be added.")
            profile[sample_name] = records[id].data.profile[sample_name]
    return profile

def extract_seqs(records):
    seqs = {}
    for id in records:
        seqs[id] = records[id].data.seq_data
    return seqs

def write_gene_fastas(seq_data,work_dir):
    d = 0
    files = {}
    for id in seq_data:
        record = seq_data[id]
        for seq_id in record:
            if 'locus_name' not in record[seq_id]:
                continue
            locus_name = record[seq_id]['locus_name']
            if locus_name == '':
                continue
            if 'dna_seq' in record[seq_id]:
                seq = record[seq_id]['dna_seq']
            else:
                seq = record[seq_id]['aa_seq']
            out_file = os.path.join(work_dir, f"{locus_name}.fas")
            if not os.path.isfile(out_file):
                oh = open(out_file,'w')
                files[locus_name] = {'file':out_file}
            else:
                if not locus_name in files:
                    oh = open(out_file,'w')
                    files[locus_name] = {'file':out_file}
                else:
                    oh = open(out_file,'a')
            seq_name = f'{id}'
            oh.write(f'>{seq_name}\n{seq}\n')
            oh.close()
            d+=1
    return files

def read_samplesheet(sample_key_file):
    """
    We will be accepting a file with two columns either ['sample', 'mlst_alleles'] or ['sample_name', 'mlst_alleles']
    """
    try:
        df = pd.read_csv(sample_key_file, sep=",", usecols= lambda column: column in {"sample", "sample_name", "mlst_alleles"}, header=0)
        df["mlst_alleles"] = df["mlst_alleles"].apply(lambda x: os.path.basename(x))
        sampledict = df.set_index("mlst_alleles").T.to_dict('list')


    except:
            logging.critical("File {fname} should be CSV that will contain either ['sample', 'mlst_alleles'] or ['sample_name', 'mlst_alleles']".format(fname = sample_key_file))
            raise FileNotFoundError("Incorrect file format")

    return sampledict

def validate_and_fix_profiles(mlst, sample_id, file_name):
    data_key = "data"
    profile_key = "profile"
    # Extract the profile from the json_data
    profile = mlst.get(data_key, {}).get(profile_key, {})
    # Check for multiple keys in the JSON file and define error message
    keys = sorted(profile.keys())
    original_key = keys[0] if keys else None
    # Define a variable to store the match_status (True or False)
    match_status = sample_id in profile
    # Initialize the error message
    MLST_message = None

    if not keys:
        logger.critical(f"{file_name} is missing the 'profile' section or is completely empty!")
        raise ValueError(f"{file_name} is missing the 'profile' section or is completely empty!")

    elif len(keys) > 1:
        # Check if sample_id matches any key
        if not match_status:
            MLST_message = f"No key in the MLST JSON file ({file_name}) matches the specified sample ID '{sample_id}'. The first key '{original_key}' has been forcefully changed to '{sample_id}' and all other keys have been removed."
            # Retain only the specified sample ID
            mlst[data_key][profile_key] = {sample_id: profile[original_key]}
        else:
            MLST_message = f"MLST JSON file ({file_name}) contains multiple keys: {keys}. The MLST JSON file has been modified to retain only the '{sample_id}' entry"
            # Retain only the specified sample_id in the profile
            mlst[data_key][profile_key] = {sample_id: profile[sample_id]}
    elif not match_status:
        MLST_message = f"{sample_id} ID and JSON key in {file_name} DO NOT MATCH. The '{original_key}' key in {file_name} has been forcefully changed to '{sample_id}': User should manually check input files to ensure correctness."
        # Update the JSON file with the new sample ID
        mlst[data_key][profile_key] = {sample_id: profile[original_key]}
        mlst[data_key]["sample_name"] = sample_id

    # Create a report for all the samples that have their profiles modified in the output profile.tsv
    if MLST_message:
        mlst_report = OutputRow(sample_id, keys, MLST_message)
    else:
        mlst_report = None

    # Write the updated JSON data back to the original file
    return mlst, mlst_report

def run_merge(config):
    analysis_parameters = config

    #Input Parameters
    input_files = config['input'][0]
    outdir = config['outdir']

    if config['loci'] is not None:
        if os.path.isfile(config['loci']):
            loci_to_keep = pd.read_csv(config['loci'], header=None).iloc[:, 0].tolist()
        else:
            loci_to_keep = config['loci'].split(',')
        print(f"INFO:locidex merge: Keeping only the following loci: {loci_to_keep}")
    else:
        loci_to_keep = config['loci']

    ###
    # Commented out as these changes will require test data
    # perform_align = config['align']
    # linker_seq = config['linker']
    # n_threads = config['n_threads']
    ###
    force = config['force']
    validate_db = config['strict']
    profile_refs = config['profile_ref']
    if validate_db is None or validate_db == '':
        validate_db = False
    if profile_refs is None or profile_refs == '':
        profile_refs = False

    run_data = {}
    run_data['analysis_start_time'] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    run_data['parameters'] = analysis_parameters

    if os.path.isdir(outdir) and not force:
        logger.critical(f'Error {outdir} exists, if you would like to overwrite, then specify --force')
        raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST), str(outdir))

    if not os.path.isdir(outdir):
        os.makedirs(outdir, 0o755)

    #perform merge
    file_list = get_file_list(input_files)
    records, modified_MLST_file_list = read_file_list(file_list,perform_db_validation=validate_db, perform_profile_validation=profile_refs, loci_to_keep=loci_to_keep)

    #create profile
    df = pd.DataFrame.from_dict(extract_profiles(records), orient='index')
    df.insert(loc=0, column='sample_id', value=df.index.tolist())
    df.to_csv(os.path.join(outdir,'profile.tsv'),index=False,header=True,sep="\t")

    del(df)
    run_data['result_file'] = os.path.join(outdir,"profile.tsv")

    #Write report of all the MLST files with profile mismatch and how MLST profiles with mismatch were modified
    if modified_MLST_file_list:
        df = pd.DataFrame(modified_MLST_file_list)
        df.to_csv(f'{outdir}/MLST_error_report.csv', index=False, header=False)


    run_data['analysis_end_time'] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    with open(os.path.join(outdir,"run.json"),'w' ) as fh:
        fh.write(json.dumps(run_data, indent=4))


def run(cmd_args=None):
    if cmd_args is None:
        parser = add_args()
        cmd_args = parser.parse_args()
    analysis_parameters = vars(cmd_args)


    config = {}
    for p in analysis_parameters:
        if not p in config:
            config[p] = analysis_parameters[p]

    run_merge(config)


# call main function
if __name__ == '__main__':
    run()





import json,os

class metadata:
    input_file = None
    required_fields = []
    data = {}
    status = True
    messages = []

    def __init__(self,input_file,required_fields):
        self.input_file = input_file
        if not os.path.isfile(self.input_file):
            self.messages.append(f'Error {self.input_file} input file does not exist')
            self.status = False
            return self.status
        self.data = self.parse_file()
        self.required_fields = required_fields
        (self.status,k) = self.validate_dict_keys(self.data,self.required_fields)
        if not self.status:
            self.messages.append(f'Error {self.input_file} input metadata file is missing needed key {k} ')


        return self.status

    def parse_file(self):
        with open(self.input_file) as file:
            self.config = json.load(file)

    def validate_dict_keys(self,data_dict, required_keys):
        for k in required_keys:
            if not k in data_dict:
                return [False, k]
        return [True, k]




"""
Tests for locidex format module
"""

import pytest
from locidex import format


import os
import json
from dataclasses import dataclass

EXPECTED_DATA_OUT = "locidex/example/format_db_mlst_out"
TEST_DATA = "locidex/example/format_db_mlst_in"

@dataclass
class FormatArgs:
    input: os.PathLike
    outdir: os.PathLike
    min_len_frac: float
    max_len_frac: float
    min_ident: float
    min_match_cov: float
    translation_table: int
    not_coding: bool
    force: bool

@pytest.fixture(scope="module")
def output_directory(tmp_path_factory):
    fn = tmp_path_factory.mktemp("build")
    return str(fn)


@pytest.fixture(scope="module")
def cmd_args(output_directory):
    command = FormatArgs(
        input=TEST_DATA,
        outdir=output_directory,
        min_len_frac=0.7,
        max_len_frac=1.3,
        min_ident=80.0,
        min_match_cov=80.0,
        translation_table=11,
        not_coding=False,
        force=True
    )
    return command


def test_format(cmd_args):
    format.run(cmd_args)

def test_outputs(output_directory):
    assert sorted(os.listdir(output_directory)) == sorted(os.listdir(EXPECTED_DATA_OUT))


def test_formatted_db_content(output_directory):
    formatted_file = "locidex.txt"
    delimiter = "\t"
    key_id_position = 2
    actual_file = os.path.join(output_directory, formatted_file)
    expected_file = os.path.join(EXPECTED_DATA_OUT, formatted_file)
    with open(actual_file, 'r') as act, open(expected_file, 'r') as expc:
        sort_key = lambda x: x.split(delimiter)[key_id_position]
        actual_text = sorted(act.readlines(), key=sort_key)
        expected_text = sorted(expc.readlines(), key=sort_key)
        for i, j in zip(actual_text, expected_text):
            #ignoring row id
            assert i.split(delimiter)[1:] == j.split(delimiter)[1:]

def test_format_results(output_directory):
    """
    Verify continuity of results.json fields
    """
    primary_key = "parameters"
    comp_fields = [
        "min_len_frac",
        "max_len_frac",
        "min_ident",
        "min_match_cov",
        "translation_table",
        "not_coding",
        "force"
    ]
    result_json = "results.json"
    expected = os.path.join(EXPECTED_DATA_OUT, result_json)
    actual = os.path.join(output_directory, result_json)
    with open(actual, 'r', encoding='utf8') as act, open(expected, 'r', encoding='utf8') as expc:
        act_js = json.load(act)[primary_key]
        expc_js = json.load(expc)[primary_key]
        for i in comp_fields:
            assert act_js[i] == expc_js[i]



"""
Test manifest module
"""

import pytest
import os
import json
import shutil
from pathlib import PosixPath, Path
from locidex import manifest
from locidex.constants import DBConfig
from dataclasses import dataclass
from typing import Dict


TEST_FAIL_AUTHOR = "locidex/example/manifest_in/fails/fails_author"
TEST_FAIL_DESC = "locidex/example/manifest_in/fails/fails_name"
TEST_PASS_MULTIPLE = "locidex/example/manifest_in/passes/pass_multiple"
TEST_PASS_SINGLE = "locidex/example/manifest_in/passes/pass_single"

@dataclass
class CMDArgs:
    input: Path


def test_db_list():
    """
    List all databases in a test directory
    """
    assert set(manifest.check_dbs(Path(TEST_PASS_MULTIPLE))) == set([PosixPath('locidex/example/manifest_in/passes/pass_multiple/pass_three_db'), 
                                                            PosixPath('locidex/example/manifest_in/passes/pass_multiple/pass_two_db'), 
                                                            PosixPath('locidex/example/manifest_in/passes/pass_multiple/pass_one_db')])

@pytest.mark.parametrize("input_dir,output",
[
    (TEST_PASS_MULTIPLE,
    [
    (PosixPath('pass_three_db'), 
    DBConfig(db_name='Locidex Database 3', db_version='1.0.0', db_date='04/04/2024', db_author='test1', db_desc='test1', db_num_seqs=53, is_nucl=True, is_prot=True, nucleotide_db_name='nucleotide', protein_db_name='protein')), 
    (PosixPath('pass_two_db'), 
    DBConfig(db_name='Locidex Database 2', db_version='1.0.0', db_date='04/04/2024', db_author='test1', db_desc='test1', db_num_seqs=53, is_nucl=True, is_prot=True, nucleotide_db_name='nucleotide', protein_db_name='protein')), 
    (PosixPath('pass_one_db'), 
    DBConfig(db_name='Locidex Database 1', db_version='1.0.0', db_date='04/04/2024', db_author='test1', db_desc='test1', db_num_seqs=53, is_nucl=True, is_prot=True, nucleotide_db_name='nucleotide', protein_db_name='protein'))]),
    (TEST_PASS_SINGLE,
    [(PosixPath('pass_one_db'), 
    DBConfig(db_name='Locidex Database', db_version='1.0.0', db_date='04/04/2024', db_author='test1', db_desc='test1', db_num_seqs=53, is_nucl=True, is_prot=True, nucleotide_db_name='nucleotide', protein_db_name='protein'))]),
])
def test_pass_validate_db_files(input_dir, output):
    input_path = Path(input_dir)
    dbs = manifest.check_dbs(input_path)
    assert sorted(manifest.validate_db_files(dbs, input_path)) == sorted(output)

def test_fail_validate_db_files_author(capsys):
    with pytest.raises(AttributeError):
        manifest.check_config(Path(TEST_FAIL_AUTHOR))
        assert "Config cannot have missing values: db_author is empty" == capsys.readouterr()

def test_fail_validate_db_files_description(capsys):
    with pytest.raises(AttributeError):
        manifest.check_config(Path(TEST_FAIL_AUTHOR))
        assert "Config cannot have missing values: db_desc is empty" == capsys.readouterr()


def test_create_manifest_multiple():
    output = {'Locidex Database 3': [
                {'path': 'pass_three_db', 
                'config': {'db_name': 'Locidex Database 3', 'db_version': '1.0.0', 'db_date': '04/04/2024', 'db_author': 'test1', 'db_desc': 'test1', 'db_num_seqs': 53, 'is_nucl': True, 'is_prot': True, 'nucleotide_db_name': 'nucleotide', 'protein_db_name': 'protein'}}], 
            'Locidex Database 2': [
                {'path': 'pass_two_db', 
                'config': {'db_name': 'Locidex Database 2', 'db_version': '1.0.0', 'db_date': '04/04/2024', 'db_author': 'test1', 'db_desc': 'test1', 'db_num_seqs': 53, 'is_nucl': True, 'is_prot': True, 'nucleotide_db_name': 'nucleotide', 'protein_db_name': 'protein'}}], 
            'Locidex Database 1': [
                {'path': 'pass_one_db', 
                'config': {'db_name': 'Locidex Database 1', 'db_version': '1.0.0', 'db_date': '04/04/2024', 'db_author': 'test1', 'db_desc': 'test1', 'db_num_seqs': 53, 'is_nucl': True, 'is_prot': True, 'nucleotide_db_name': 'nucleotide', 'protein_db_name': 'protein'}}]}
    assert manifest.create_manifest(Path(TEST_PASS_MULTIPLE)) == output

def test_create_manifest_single():
    output = {'Locidex Database': [
                {'path': 'pass_one_db', 
                'config': {'db_name': 'Locidex Database', 'db_version': '1.0.0', 'db_date': '04/04/2024', 'db_author': 'test1', 'db_desc': 'test1', 'db_num_seqs': 53, 'is_nucl': True, 'is_prot': True, 'nucleotide_db_name': 'nucleotide', 'protein_db_name': 'protein'}}] }
    assert manifest.create_manifest(Path(TEST_PASS_SINGLE)) == output


def test_write_manifest(tmpdir):
    outdir = tmpdir / "build"
    shutil.copytree(TEST_PASS_MULTIPLE, outdir)
    cmd_args = CMDArgs(input=outdir)
    file_out = manifest.run(cmd_args=cmd_args)
    assert file_out.exists()


def test_read_manifest(tmpdir):
    outdir = Path(tmpdir / "build")
    shutil.copytree(TEST_PASS_MULTIPLE, outdir)
    cmd_args = CMDArgs(input=outdir)
    file_out = manifest.run(cmd_args=cmd_args)
    assert file_out.exists()
    output = {'Locidex Database 3': [
                {'path': 'pass_three_db', 
                'config': {'db_name': 'Locidex Database 3', 'db_version': '1.0.0', 'db_date': '04/04/2024', 'db_author': 'test1', 'db_desc': 'test1', 'db_num_seqs': 53, 'is_nucl': True, 'is_prot': True, 'nucleotide_db_name': 'nucleotide', 'protein_db_name': 'protein'}}], 
            'Locidex Database 2': [
                {'path': 'pass_two_db', 
                'config': {'db_name': 'Locidex Database 2', 'db_version': '1.0.0', 'db_date': '04/04/2024', 'db_author': 'test1', 'db_desc': 'test1', 'db_num_seqs': 53, 'is_nucl': True, 'is_prot': True, 'nucleotide_db_name': 'nucleotide', 'protein_db_name': 'protein'}}], 
            'Locidex Database 1': 
                [ {'path': 'pass_one_db', 
                'config': {'db_name': 'Locidex Database 1', 'db_version': '1.0.0', 'db_date': '04/04/2024', 'db_author': 'test1', 'db_desc': 'test1', 'db_num_seqs': 53, 'is_nucl': True, 'is_prot': True, 'nucleotide_db_name': 'nucleotide', 'protein_db_name': 'protein'}}]}

    manifest_data: Dict[str, manifest.ManifestItem] = manifest.read_manifest(outdir)
    for k, v in manifest_data.items():
        comp_data = output[k]
        v = [i.to_dict() for i in v]
        assert v == comp_data


def test_select_db(tmpdir):
    outdir = Path(tmpdir / "build")
    shutil.copytree(TEST_PASS_MULTIPLE, outdir)
    cmd_args = CMDArgs(input=outdir)
    file_out = manifest.run(cmd_args=cmd_args)
    assert file_out.exists()
    db_out =  {'path': 'pass_three_db', 
                'config': {'db_name': 'Locidex Database 3', 'db_version': '1.0.0', 'db_date': '04/04/2024', 'db_author': 'test1', 'db_desc': 'test1', 'db_num_seqs': 53, 'is_nucl': True, 'is_prot': True, 'nucleotide_db_name': 'nucleotide', 'protein_db_name': 'protein'}}
    manifest_data: Dict[str, manifest.ManifestItem] = manifest.read_manifest(outdir)
    assert manifest.select_db(manifest_data, "Locidex Database 3", "1.0.0").to_dict() == db_out



from Bio import SeqIO
import os
import gzip
from mimetypes import guess_type
from functools import partial
import pyrodigal

class gene_prediction:
    file = None
    sequences = {}
    genes = []
    status = True
    messages = []

    def __init__(self,file):
        self.file = file
        if not os.path.isfile(self.file):
            self.messages.append(f'Error {self.file} does not exist')
            self.status = False


    def predict(self,num_threads=1):
        encoding = guess_type(self.file)[1]
        _open = partial(gzip.open, mode='rt') if encoding == 'gzip' else open
        with _open(self.file) as f:
            for record in SeqIO.parse(f, 'fasta'):
                orf_finder = pyrodigal.GeneFinder(meta=True)
                genes =  orf_finder.find_genes(bytes(record.seq))
                for i, pred in enumerate(genes):
                    id = f">{record.id}_{i + 1}"
                    self.sequences[id] = pred.sequence()
                self.genes.append(genes)




import json
import os
import pathlib


class db_config:
    config = {}
    input_file = None
    required_fields = []
    status = True
    messages = []

    def __init__(self,input_file,required_fields):
        self.input_file = input_file
        if not os.path.isfile(self.input_file):
            self.messages.append(f'Error {self.input_file} input file does not exist')
            self.status = False
        self.parse_file()
        self.required_fields = required_fields
        (self.status,k) = self.validate_dict_keys(self.config, self.required_fields)
        if k != '':
            self.messages.append(f'Error required field missing from config file key:{k}')
            self.status = False

        return

    def parse_file(self):
        with open(self.input_file) as file:
            self.config = json.load(file)


    def validate_dict_keys(self,data_dict, required_keys):
        for k in required_keys:
            if not k in data_dict:
                return [False, k]
        return [True, '']


class search_db_conf:
    input_dir = None
    files = {
        'dir':[],
        'file':[]
    }
    config_file_path = ''
    config_obj = None
    meta_file_path = ''
    blast_paths = {

    }

    status = True
    messages = []

    def __init__(self, input_dir,db_basenames,config_required_fields,):
        self.input_dir = input_dir
        if not os.path.isdir(self.input_dir):
            self.messages.append(f'Error {self.input_dir} input file does not exist')
            self.status = False
            return
        self.get_dir_files()

        p = self.get_file_path_by_name(db_basenames['config'])

        if p == '':
            conf_name = db_basenames['config']
            self.messages.append(f'Error missing expected config file {conf_name} in db {input_dir}')
            self.status = False
            return

        self.config_file_path = p
        self.config_required_fields = config_required_fields
        self.config_obj = db_config(self.config_file_path,self.config_required_fields)
        self.status = self.config_obj.status
        self.messages = self.messages + self.config_obj.messages

        p = self.get_file_path_by_name(db_basenames['meta'])
        if p == '':
            conf_name = db_basenames['meta']
            self.messages.append(f'Error missing expected meta file {conf_name} in db {input_dir}')
            self.status = False
            return

        self.meta_file_path = p

        p = self.get_file_path_by_name('blast','dir')

        if self.config_obj.config["is_nucl"]:
            self.blast_paths['nucleotide'] = os.path.join(p,"nucleotide/nucleotide")
        if self.config_obj.config["is_prot"]:
            self.blast_paths['protein'] = os.path.join(p, "protein/protein")


        if len(self.blast_paths) == 0:
            self.messages.append(f'Error could not find any expected blast dbs in path: {input_dir}, expected {db_basenames}')
            self.status = False
            return


    def get_dir_files(self):
        d = pathlib.Path(self.input_dir).rglob('*')
        for item in d:
            type = 'file'
            if item.is_dir():
                type = 'dir'
            self.files[type].append([f"{item.resolve()}",os.path.basename(item)])

    def get_file_path_by_name(self,filename,t='file'):
        for f in self.files[t]:
            if filename == f[1]:
                return f[0]
        return ''








#!/usr/bin/env python3

import sys
import argparse
import traceback
import logging
from . import format, extract, report, merge, search, build, manifest, utils, constants

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.DEBUG)

tasks = {
    'search': (search, 'Query set of Loci/Genes against a database to produce a sequence store for downstream processing'),
    'extract': (extract, 'Extract loci from a genome based on a locidex database'),
    'report': (report, 'Filter a sequence store and produce and extract of blast results and gene profile'),
    'merge':  (merge, 'Merge a set of gene profiles into a standard profile format'),
    'format': (format, 'Format fasta files from other MLST databases for use with locidex build'),
    'build': (build, 'Build a locidex database'),
    'manifest': (manifest, 'Create a multi-database folder manifest'),
}


def main(argv=None):
    module_idx = 0
    help_msg = 1
    parser = argparse.ArgumentParser(prog="locidex")
    sub_parsers = parser.add_subparsers(dest="command")
    for k, v in tasks.items():
        format_parser = sub_parsers.add_parser(k, description=v[help_msg], help=v[help_msg])
        v[module_idx].add_args(format_parser)
    
    args = parser.parse_args(argv)
    if args.command is None:
        parser.print_help()
        try:
            utils.check_utilities(logger, constants.UTILITIES_CHECK)
        except FileNotFoundError:
            pass
        sys.exit()

    error_file = "errors.txt"
    try:
        utils.check_utilities(logger, constants.UTILITIES_CHECK)
        logger.info("Running {}".format(args.command))
        tasks[args.command][module_idx].run(args)
        logger.info("Finished: {}".format(args.command))
    except Exception as e:
        with open(error_file, "w") as f:
            f.write(traceback.format_exc())
        error_number = e.errno if hasattr(e, "errno") else -1
        logger.critical("Program exited with errors, please review logs. For the full traceback please see file: {}".format(error_file))
        SystemExit(error_number)
    else:
        logger.info("Program finished without errors.")

# call main function
if __name__ == '__main__':
    main()


import sys

from locidex.classes.mafft import mafft
from multiprocessing import Pool, cpu_count
import os
from locidex.classes import run_command


def stage_data(seq_data,work_dir):
    d = 0
    for id in seq_data:
        record = seq_data[id]
        locus_name = record['locus_name']
        ref_id = record['ref_id']
        ref_seq = record['ref_seq']
        ext_seq = record['ext_seq']
        out_file = os.path.join(work_dir, f"{d}-{locus_name}.fas")
        write_seq({ref_id: ref_seq, id: ext_seq}, out_file)
        seq_data[id]['file'] = out_file
        d += 1


def write_seq(data, out_file):
    with open(out_file, 'w') as oh:
        for id in data:
            oh.write(">{}\n{}\n".format(id, data[id]))



def align(input_fasta,params={'globalpair':''}):
    command = ['mafft']
    for p in params:
        command.append(f'--{p}')
        command.append(params[p])
    command.append(input_fasta)
    command = " ".join([str(x) for x in command])
    (stdout, stderr) = run_command(command)
    return (stdout,stderr)


def parse_align(align):
    align = align.split('\n')
    seqs = {}
    for row in align:
        if len(row) == 0:
            continue
        if row[0] == '>':
            id = row[1:]
            seqs[id] = []
            continue
        seqs[id].append(row.lower())
    for id in seqs:
        seqs[id] = "".join([str(x) for x in seqs[id]])
    return seqs

def perform_alignment(seq_data,work_dir,num_threads=1):
    try:
        sys_num_cpus = len(os.sched_getaffinity(0))
    except AttributeError:
        sys_num_cpus = cpu_count()

    if num_threads > sys_num_cpus:
        num_threads = sys_num_cpus

    stage_data(seq_data, work_dir)

    pool = Pool(processes=num_threads)

    results = []
    for id in seq_data:
        record = seq_data[id]
        results.append(pool.apply_async(align, args=((record['file'],))))

    pool.close()
    pool.join()

    r = []
    for x in results:
        if isinstance(x, dict):
            r.append(x)
        else:
            r.append(x.get())

    ids = list(seq_data.keys())
    for i in range(0, len(r)):
        id = ids[i]
        if len(r[i]) == 0:
            continue
        seq_data[id]['alignment'] = parse_align(r[i][0])

    return r


class aligner:
    def __init__(self,trim_fwd=True,trim_rev=True,ext_fwd=True, ext_rev=True,fill=True,snps_only=False):
        self.trim_fwd = trim_fwd
        self.trim_rev = trim_rev
        self.ext_fwd = ext_fwd
        self.ext_rev = ext_rev
        self.fill = fill
        self.snps_only = snps_only


    def transform_seq(self,seq1,seq2,trim_fwd=True,trim_rev=True,ext_fwd=True, ext_rev=True,gap_fill=True,snps_only=False):
        if snps_only:
            length = len(seq1)
            tseq = list(seq1)
            for i in range(0,length):
                b1 = seq1[i]
                b2 = seq2[i]
                if b1 != '-' and b2 != '-':
                    tseq[i] = b2

            tseq = "".join([str(x) for x in tseq])
        else:
            tseq = self.trim_ends(seq1,seq2,trim_fwd,trim_rev)
            tseq = self.extend(seq1, tseq, ext_fwd, ext_rev)
            if gap_fill:
                tseq = self.gap_fill(seq1, tseq)
        return tseq


    def trim_ends(self,seq1,seq2,trim_fwd=True,trim_rev=True):
        seq = list(seq2)
        is_end = False
        length = len(seq1)
        if trim_fwd:
            for i in range(0,length):
                b1 = seq1[i]
                if b1 == '-':
                    is_end = True
                if is_end and b1 == '-':
                    seq[i] = '-'
                else:
                    is_end = False
                    break

        is_end = False
        length = len(seq1)
        if trim_rev:
            for i in reversed(range(0,length)):
                b1 = seq1[i]
                if b1 == '-':
                    is_end = True
                if is_end and b1 == '-':
                    seq[i] = '-'
                else:
                    break
        return "".join([str(x) for x in seq])



    def extend(self, seq1, seq2,ext_fwd=True, ext_rev=True):
        seq = list(seq2)
        is_end = True
        length = len(seq1)

        if ext_fwd:
            for i in range(0, length):
                b1 = seq1[i]
                b2 = seq2[i]
                if b1 == '-':
                    is_end = False
                if is_end and b1 != '-' and b2 == '-':
                    seq[i] = b1


        is_end = True
        length = len(seq1)
        if ext_rev:
            for i in reversed(range(0, length)):
                b1 = seq1[i]
                b2 = seq2[i]
                if b1 == '-':
                    is_end = False

                if is_end and b1 != '-'  and b2 == '-':
                    seq[i] = b1

        return "".join([str(x) for x in seq])

    def gap_fill(self, seq1, seq2):
        seq = []
        is_end = True
        length = len(seq1)
        for i in range(0, length):
            b1 = seq1[i]
            b2 = seq2[i]
            if b1 == '-':
                is_end = False
            if is_end and b2 == '-':
                b2 = b1
            seq.append(b2)
        return "".join([str(x) for x in seq])

    def calc_coverage(self,seq1,seq2):
        coverage = 0
        total = 0
        for idx,base in enumerate(seq1):
            b1 = seq1[idx]
            b2 = seq2[idx]
            if b1 != '-':
                total+=1
                if b2 != '-':
                    coverage+=1
        if total > 0:
            return coverage / total * 100
        return 0


    def call_seq(self,id, ref_seq, query_seq):
        seq_record = {
            'id': id,
            'matched_bases': 0,
            'num_diff': 0,
            'coverage': 0,
            'seq': ''
        }
        if self.trim_fwd or self.trim_rev or self.ext_fwd or self.ext_rev or self.gap_fill:
            query_seq = self.transform_seq(ref_seq,query_seq,self.trim_fwd,self.trim_rev,
                                      self.ext_fwd, self.ext_rev,self.fill,self.snps_only)
        (match, diff) = self.count_identity(ref_seq,query_seq)
        seq_record['matched_bases'] = match
        seq_record['num_diff'] = diff
        seq_record['coverage'] = self.calc_coverage(ref_seq,query_seq)
        seq_record['seq'] = query_seq
        return seq_record

    def count_identity(self,seq1,seq2):
        match = 0
        diff = 0
        for idx, base in enumerate(seq1):
            if seq1[idx] != '-' and seq1[idx] == seq2[idx]:
                match+=1
            else:
                diff+=1
        return [match, diff]



import pytest
import os
import locidex.classes.aligner
from locidex.constants import BlastColumns
from dataclasses import dataclass

PACKAGE_ROOT = os.path.dirname(locidex.__file__)
@dataclass
class Datasets:
    def __init__(self, tmpdir):
        self.align_inputs = {'thrA':{'thrA':{
                                'locus_name': 'thrA', 
                                'ref_id': '43', 
                                'ref_seq': 'gtgctgggccgtaatggttccgactattccgccgccgtgctggccgcctgtttacgcgctgactgctgtgaaatctggactgacgtcgatggcgtgtatacctgtgacccgcgccaggtgccggacgccagactgctgaaatcgatgtcctaccaggaagcgatggaactctcttacttcggcgccaaagtccttcaccctcgcaccataacgcctatcgcccagttccagatcccctgtctgattaaaaataccggtaatccgcaggcgccaggaacgctgatcggcgcgtccagcgacgatgataatctgccggttaaagggatctctaaccttaacaacatggcgatgtttagcgtctccggcccgggaatgaaagggatgattgggatggcggcgcgtgttttcgccgccatgtctcgcgccgggatctcggtggtgctcattacccagtcctcctctgagtacagcatcagcttctgtgtgccgcagagtgactgc', 
                                'ext_seq': 'gtgctgggccgtaatggttccgactattccgccgccgtgctggccgcctgtttacgcgctgactgctgtgaaatctggactgacgtcgatggcgtgtatacctgtgacccgcgccaggtgccggacgccaggctgttgaaatcgatgtcctaccaggaagcgatggagctctcttacttcggcgctaaagtccttcaccctcgcaccataacgcctatcgcccagttccagatcccctgtctgattaaaaataccggcaatccgcaggcgccaggaacgctgatcggcgcgtccagcgacgatgataatctgccggttaaagggatctctaaccttaacaacatggcgatgtttagcgtctccggcccgggaatgaaagggatgattgggatggcggcgcgtgttttcgccgccatgtctcgcgccgggatctcggtggtgctcattacccagtcctcctctgagtacagcatcagcttctgtgtgccgcagagtgactgc', 
                                'file': tmpdir
                            }},
                            'hemD':{'hemD': {
                                'locus_name': 'hemD', 
                                'ref_id': '16', 
                                'ref_seq': 'gcgacgttgacggaaaacgatctggtttttgccctttcacagcacgccgtcgcctttgcccacgcccaactccagcgagatggtcgaaactggcctgcgtcgccgcgctatttcgcgattggtcgcaccacggcgctcgcccttcataccgttagcgggttcgatattcgttatccattggatcgggaaatcagcgaagtcttgctacaattacctgaattacaaaatattgcgggcaaacgcgcgctgattttgcgtggcaatggcggtcgcggtcgcgaactgctgggcgaaaccctgacagctcgcggagccgaagtcagtttttgtgaatgttatcaacgaagtgcgaaacattacgatggcgcagaagaggcgatgcgctggcacactcgcggcgtaacgacgcttgttgtcaccagcggcgagatgttgcaa', 
                                'ext_seq': 'gcaacgctgacggaaaacgatctggtttttgccctttcacagcactccgtcgcctttgctcacgcccagctccagcgggatggacgaaactggcctgcgtcgccgcgctatttctcgattggccgcaccacggcgctcgcccttcataccgttagcgggttcgatattcgttatccattggatcgggaaatcagcgaagccttgctacaattacctgaattacaaaatattgcgggcaaacgcgcgctgattttgcgtggcaatggcggccgcgaactgctgggcgaaaccctgacagttcgcggagccgaagtcagtttttgtgaatgttatcaacgatgtgcgaaacattacgatggcgcggaagaagcgatgcgctggcatactcgcggcgtaacaacgcttgttgttaccagcggcgagatgttgcaa', 
                                'file': tmpdir
                            }}
                            }

@pytest.fixture
def aligner_class_fixture():
    return locidex.classes.aligner.aligner(trim_fwd=True,trim_rev=True,ext_fwd=False, ext_rev=False,fill=False, snps_only=False)

def test_perform_alignment_call_variants(tmpdir, aligner_class_fixture):
    seq_id = 'thrA'; ref_id='43'
    ext_seq_data=Datasets(tmpdir).align_inputs[seq_id]     
    result = locidex.classes.aligner.perform_alignment(ext_seq_data, tmpdir, 1)
    alignment = ext_seq_data[seq_id]['alignment']
    variants = aligner_class_fixture.call_seq(seq_id,alignment[ref_id],alignment[seq_id])
    assert list(variants.keys())==['id', 'matched_bases', 'num_diff', 'coverage', 'seq']
    assert variants['num_diff'] == 5, 'Expected value 5 for sequence differences'
    assert variants['coverage'] == 100, 'Expected value 100 for reference coverage'

    aligned_sequences_str=result[0][0] #identical to alignment variable, but needed to test parser
    aligned_sequences_dict = locidex.classes.aligner.parse_align(aligned_sequences_str)
    assert list(aligned_sequences_dict.keys()) == [ref_id, seq_id]
    assert all([True if 'alignment' in ext_seq_data[key] else False for key in ext_seq_data ]), 'Missing alignment results in locidex.classes.aligner object'
    assert len(ext_seq_data[seq_id]['alignment'].keys()) == 2, 'Should have two sequences (ref and query)'
    assert len(ext_seq_data[seq_id]['alignment'][ref_id]) == 501, 'Length should be 501 bp'
    assert len(ext_seq_data[seq_id]['alignment'][seq_id]) == 501, 'Length should be 501 bp'
    assert list(ext_seq_data[seq_id].keys()) == ['locus_name', 'ref_id', 'ref_seq', 'ext_seq', 'file', 'alignment'], 'Keys do not match'

def test_perform_seq_transform_trim_extend_gapfill(aligner_class_fixture):
    tseq = aligner_class_fixture.transform_seq( 'GTGGCAATGGCGGT',
                                                'GCGTGGCAATG---')
    assert tseq == 'GCGTGGCAATGGGT', f'Unexpected result {tseq}'
    tseq = aligner_class_fixture.trim_ends('ATGT--','ATGTAA', trim_fwd=False,trim_rev=False)
    assert 'ATGTAA' == tseq, f'Unexpected result {tseq}'
    tseq = aligner_class_fixture.trim_ends('--GTAA','ATGTAA', trim_fwd=False,trim_rev=True)
    assert 'ATGTAA' == tseq, f'Unexpected result {tseq}'
    tseq = aligner_class_fixture.trim_ends('--GT--','ATGTAA', trim_fwd=True,trim_rev=True)
    assert '--GT--' == tseq, f'Unexpected result {tseq}'
    tseq = aligner_class_fixture.extend('GCGTAA','-AGTCC',ext_fwd=False, ext_rev=False)
    assert '-AGTCC' == tseq, f'Unexpected result {tseq}'
    tseq = aligner_class_fixture.extend('GCGTAA','GAGTC-',ext_fwd=False, ext_rev=True)
    assert 'GAGTCA' == tseq, f'Unexpected result {tseq}'
    tseq = aligner_class_fixture.extend('GCGTAA','-AGTCC',ext_fwd=True, ext_rev=True)
    assert 'GAGTCC' == tseq, f'Unexpected result {tseq}'
    tseq = aligner_class_fixture.gap_fill('ATGTAA','AT--AA')
    assert 'ATGTAA' == tseq, f'Unexpected result {tseq}'

def test_coverage_and_identity(aligner_class_fixture):
    coverage = aligner_class_fixture.calc_coverage('GCGTAACCAT','GCGTAACC--')
    assert coverage == 80, "Expected different coverage value between two sequences"
    coverage = aligner_class_fixture.calc_coverage('GCGTAACCAT','-CGTA-CCA-')
    assert coverage == 70, "Expected different coverage value between two sequences"
    identity = aligner_class_fixture.count_identity('GCGTAACCAT','GCGTAACC--')
    assert identity == [8,2], "Expected different match and difference value in list"
    identity = aligner_class_fixture.count_identity('GCGTAACCAT','-CGTAACCGG')
    assert identity == [7,3], "Expected different match and difference value in list"


from subprocess import Popen, PIPE
def run_command(command):
    """
    This method runs the passed command on the shell command line.

    Parameters
    ----------
    command : str
        The command for the command line

    Returns
    -------
    bytes, bytes
        stdout, stderr: the standard out and error messages returned by the command line
    """
    p = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate()
    stdout = stdout.decode('utf-8')
    stderr = stderr.decode('utf-8')
    return stdout, stderr



import pytest
import pandas as pd
from locidex.classes.assignment import assignment

# Fixture to create a sample DataFrame
@pytest.fixture
def sample_dataframe():
    data = {
        'query_id': [1, 2, 2, 3],
        'value': [100, 200, 200, 300],
        'another_column': ['A', 'B', 'B', 'C']
    }
    df = pd.DataFrame(data)
    loci_metadata = {'locus_1': 'metadata_1', 'locus_2': 'metadata_2'}
    return df, loci_metadata

def test_assignment_initialization(sample_dataframe):
    df, loci_metadata = sample_dataframe
    query_id_col = 'query_id'
    assgn = assignment(df, query_id_col, loci_metadata)
    
    # Check if the DataFrame is correctly set
    assert assgn.df.equals(df), "DataFrame not correctly assigned."
    
    # Check if the columns attribute correctly lists the DataFrame's columns
    expected_columns = ['query_id', 'value', 'another_column']
    assert assgn.columns == expected_columns, "Columns attribute not correctly set."
    
    # Check if loci_metadata is correctly set
    expected_loci_metadata = {'locus_1': 'metadata_1', 'locus_2': 'metadata_2'}
    assert assgn.loci_metadata == expected_loci_metadata, "Loci metadata not correctly assigned."
    
    # Check if the unique queries list is correctly set
    expected_queries = [1, 2, 3]
    assert assgn.queries == expected_queries, "Queries list not correctly set."


import pytest
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from locidex.classes.mafft import mafft

@pytest.fixture
def temp_fasta(tmp_path):
    # Fixture to create a temporary FASTA file with predefined sequences.
    records = [
        SeqRecord(Seq("ATGCATGCATGC"), id="seq1", description=""),
        SeqRecord(Seq("GTACGTACGTAC"), id="seq2", description="")
    ]

    # Create a temporary FASTA file
    fasta_path = tmp_path / "example.fasta"
    SeqIO.write(records, str(fasta_path), "fasta")

    # Return the path to the temporary FASTA file
    return str(fasta_path)

def test_mafft_alignment(temp_fasta):
    # Test the functionality of the mafft class using the temporary FASTA file.
    params = {"thread": "1"}
    mafft_instance = mafft(input_fasta=temp_fasta, params=params)
    assert mafft_instance.status, f"MAFFT did not execute successfully: {mafft_instance.messages}"
    alignment = mafft_instance.get_alignment()
    assert "seq1" in alignment['alignment'], "seq1 not found in alignment"
    assert "seq2" in alignment['alignment'], "seq2 not found in alignment"
    assert len(alignment['alignment']['seq1']) > 0, "Alignment for seq1 is empty"
    assert len(alignment['alignment']['seq2']) > 0, "Alignment for seq2 is empty"

def test_mafft_failure():
    # Test the mafft class with a non-existent input file to handle failure.
    params = {"thread": "1"}
    mafft_instance = mafft(input_fasta="nonexistent.fasta", params=params)
    assert not mafft_instance.status, "MAFFT did not handle non-existent file correctly"
    expected_error_substring = "Cannot open nonexistent.fasta"
    error_found = any(expected_error_substring in message for message in mafft_instance.messages)
    assert error_found, f"Expected error message substring '{expected_error_substring}' not found in messages: {mafft_instance.messages}"


import json
import pytest
import tempfile
import os
from pathlib import Path
from locidex.classes.db import db_config, search_db_conf
import locidex.constants as constants 

required_fields = ["database", "user", "password"]

# Fixture for creating and cleaning up a temporary configuration file
@pytest.fixture
def temp_config():
    temp_file = tempfile.NamedTemporaryFile(delete=False, mode='w')
    yield temp_file.name  # Provide the path to the test
    os.unlink(temp_file.name)  # Cleanup

def write_temp_config(path, data):
    with open(path, 'w') as tmp:
        json.dump(data, tmp)

def test_db_config_valid(temp_config):
    valid_config_data = {
        "database": "test_db",
        "user": "test_user",
        "password": "test_pass"
    }
    write_temp_config(temp_config, valid_config_data)
    config = db_config(temp_config, required_fields)
    assert config.status is True
    assert config.messages == []

def test_db_config_empty_file(temp_config):
    # Assuming that the locidex errors out with Json decode error if JSON file is empty:
    with pytest.raises(json.JSONDecodeError):
        db_config(temp_config, required_fields)

    # config = db_config(temp_config, required_fields)
    # assert config.status is False
    # assert "required field missing" in config.messages[0]

def test_db_config_no_required_fields_present(temp_config):
    no_required_fields_data = {"unrelated_field": "some_value"}
    write_temp_config(temp_config, no_required_fields_data)
    config = db_config(temp_config, required_fields)
    assert config.status is False
    assert "required field missing" in config.messages[0]

def test_db_config_initialization(temp_config):
    # Create a configuration with all required fields present
    config_data = {"database": "test_db", "user": "test_user", "password": "test_pass"}
    write_temp_config(temp_config, config_data)
    
    config = db_config(temp_config, required_fields)
    
    assert config.status is True, "Initialization failed when it should have succeeded."
    assert config.config == config_data, "Configuration data was not correctly parsed."

def test_validate_dict_keys():
    # Create a minimal but valid configuration file
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as tmp:
        # Assuming an empty JSON object is a valid configuration for your class
        json.dump({}, tmp)
        tmp_path = tmp.name

    # Now that we have a valid (though minimal) configuration file, proceed with testing
    try:
        config = db_config(tmp_path, required_fields=[])

        # Test validation logic
        data_dict = {"database": "db", "user": "usr", "password": "pwd"}
        status, missing_key = config.validate_dict_keys(data_dict, ["database", "user", "password"])
        assert status is True, "Failed to validate correctly when all keys are present."
        assert missing_key == '', "Reported a missing key when none was missing."

        status, missing_key = config.validate_dict_keys(data_dict, ["missing_key"])
        assert status is False, "Failed to identify missing key."
        assert missing_key == "missing_key", "Identified the wrong missing key."
    finally:
        # Cleanup - remove the temporary file
        os.remove(tmp_path)

## search db config tests

@pytest.fixture
def temp_db_dir(tmp_path):
    # Create a temporary directory structure for testing
    db_dir = tmp_path / "db"
    db_dir.mkdir()
    config_file = db_dir / "db_config.json"
    meta_file = db_dir / "db_meta.json"
    blast_dir = db_dir / "blast"
    blast_dir.mkdir()
    nucleotide_dir = blast_dir / "nucleotide"
    nucleotide_dir.mkdir()
    protein_dir = blast_dir / "protein"
    protein_dir.mkdir()
    
    # Create mock files
    config_file.write_text(json.dumps({i: "value" for i in constants.DBConfig._keys()}))
    meta_file.write_text(json.dumps({"meta": "data"}))
    (nucleotide_dir / "nucleotide").touch()
    (protein_dir / "protein").touch()
    
    yield db_dir

def test_search_db_conf_initialization_and_blast_paths_setup(temp_db_dir):
    # Given that the intialization of the class performs setup operations, the init and blast path setup were tested together
    # Setup basenames and required fields
    db_basenames = {
        "config": "db_config.json",
        "meta": "db_meta.json"
    }
    #required_fields = ["key"]
    #required_fields = [*constants.DB_CONFIG_FIELDS]
    required_fields = [*constants.DBConfig._keys()]
    
    # Initialize search_db_conf
    search_conf = search_db_conf(str(temp_db_dir), db_basenames, required_fields)
    
    # Assertions to verify initialization was successful
    assert search_conf.status is True, "Failed to initialize with valid directory and files"
    assert search_conf.config_file_path == str(temp_db_dir / "db_config.json"), "Config file path incorrect"
    assert search_conf.meta_file_path == str(temp_db_dir / "db_meta.json"), "Meta file path incorrect"

    # Further assertions to verify blast paths setup as part of the initialization
    assert search_conf.blast_paths['nucleotide'] == str(temp_db_dir / "blast/nucleotide/nucleotide"), "Nucleotide BLAST path incorrect"
    assert search_conf.blast_paths['protein'] == str(temp_db_dir / "blast/protein/protein"), "Protein BLAST path incorrect"



from Bio import SeqIO
import gzip
from mimetypes import guess_type
from functools import partial
import os
from locidex.utils import calc_md5, slots
import logging
from locidex.constants import raise_file_not_found_e
from dataclasses import dataclass
from pathlib import Path
import sys

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.DEBUG)

@dataclass
class Fasta:
    """
    """
    gene_name: str
    seq_id: str
    seq: str
    hash: str
    length: int
    __slots__ = slots(__annotations__)


class ParseFasta:

    def __init__(self, input_file: Path,parse_def=False,seq_type=None,delim="|"):
        self.input_file = input_file
        if not self.input_file.exists():
            raise_file_not_found_e(self.input_file, logger)

        self.delim = delim
        self.seq_type = seq_type
        self.parse_def = parse_def
        self.seq_obj = self.parse_fasta()

    @staticmethod
    def normalize_sequence(fasta:str) -> str:
        """
        Remove INDELS and lower all characters in sequence
        """
        return fasta.lower().replace("-", "")

    def get_seqids(self):
        if self.seq_obj:
            return list(self.seq_obj.keys())
        raise AssertionError("No fasta file loaded.")

    def get_seq_by_id(self, fasta_id: str):
        if seq_data := self.seq_obj.get(fasta_id): #is not None and id in self.seq_obj:
            return seq_data
        raise KeyError("Missing sequence id: {}".format(fasta_id))

    def parse_fasta(self):
        encoding = guess_type(self.input_file)[1]
        _open = partial(gzip.open, mode='rt') if encoding == 'gzip' else open
        data = {}
        with _open(self.input_file) as f:
            for record in SeqIO.parse(f, 'fasta'):
                id = str(record.id)
                seq = str(record.seq)
                gene_name = id
                seq_id = id
                if self.parse_def:
                    h = id.split(self.delim)
                    if len(h) > 1:
                        gene_name = h[0]
                        seq_id = h[1]
                data[id] = Fasta(gene_name=gene_name, seq_id=seq_id, seq=self.normalize_sequence(seq), hash=calc_md5([seq])[0], length=len(seq))
        return data




class assignment:
    df = None
    columns = []
    queries = []
    loci_metadata = {}
    status = True
    messages = []

    def __init__(self, df,query_id_col,loci_metadata):
        self.df = df
        self.columns = df.columns.tolist()
        self.loci_metadata = loci_metadata
        self.queries = df[query_id_col].unique().tolist()






"""
Blast module refactored
"""

import pandas as pd

from locidex.classes import run_command
from locidex.utils import slots
from locidex.manifest import DBData
from locidex.constants import BlastCommands
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, List, Dict
import logging
import sys

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.DEBUG)

@dataclass
class FilterOptions:
    min: Optional[float]
    max: Optional[float] 
    include: Optional[bool] 
    __slots__ = slots(__annotations__)


class BlastMakeDB:
    """
    Create a blast database
    """

    def __init__(self, input_file: Path, db_type: str, parse_seqids: bool, output_db_path: Optional[Path]):
        self.input_file = input_file
        self.db_type = db_type
        self.parse_seqids = parse_seqids
        self.output_db_path = output_db_path

    def makeblastdb(self):
        """
        Create a blast datbase output
        """
        command = ['makeblastdb', '-in', str(self.input_file), '-dbtype', self.db_type]
        if self.parse_seqids:
            command.append('-parse_seqids')
        if self.output_db_path != None:
            command +=['-out', str(self.output_db_path)]
        mk_db_cmd = " ".join([str(x) for x in command])
        logger.info("Blast database command: {}".format(mk_db_cmd))
        stdout, stderr = run_command(mk_db_cmd)
        if stdout:
            logger.debug("Blast makedb stdout: {}".format(stdout))
        if stderr:
            logger.debug("Blast makedb stderr: {}".format(stderr))
        
        self.checkblastdb(self.output_db_path)
        return self.output_db_path
    
    def checkblastdb(self, blast_db_path):
        """
        Collect log information about blast database generated.
        """
        command = " ".join(["blastdbcheck", '-dbtype', str(self.db_type), '-db', str(blast_db_path)])
        logger.info("Checking generated blast database with command: {}".format(command))
        stdout, stderr = run_command(command)
        if stdout:
            logger.debug("Check Blast DB commands: {}".format(stdout))
        if stderr:
            logger.debug("Check Blast DB stderr: {}".format(stderr))


class BlastSearch:
    __blast_commands = set(BlastCommands._keys())
    __blast_extensions_nt = frozenset(['.nsq', '.nin', '.nhr'])
    __blast_extensions_pt = frozenset(['.pto', '.ptf', '.phr'])
    __filter_columns = ["qseqid", "sseqid"]

    def __init__(self, db_data: DBData, query_path: Path, blast_params: dict, blast_method: str, blast_columns: List[str], filter_options: Optional[Dict[str, FilterOptions]]=None):
        self.db_data = db_data
        if blast_method not in self.__blast_commands:
            raise ValueError("{} is not a valid blast command please pick from {}".format(blast_method, self.__blast_commands))
        self.query_path = query_path
        self.blast_params = blast_params
        self.blast_method = blast_method
        self.blast_columns = blast_columns
        self.filter_options = filter_options

    def get_blast_data(self, db_path: Path, output: Path) -> pd.DataFrame:
        """
        Run blast and parse results
        TODO need to clean up the db_path hand off from the DBData obj its dirty
        """
        
        stdout, stderr = self._run_blast(db=db_path, output=output)
        if stdout:
            logger.debug("Blast search stdout: {}".format(stdout))
        if stderr:
            logger.debug("Blast search stderr: {}".format(stderr))

        blast_data = self.parse_blast(output_file=output)
        return blast_data

    def parse_blast(self, output_file: Path):
        """
        Parse a blast output file
        output_file Path: Generate blast data
        """
        df = self.read_hit_table(output_file)
        columns = df.columns.tolist()
        for id_col in self.__filter_columns:
            tp = {}
            if id_col in columns:
                tp[id_col] = 'object'
            df = df.astype(tp)
        if self.filter_options is None:
            return df
        
        for col_name in self.filter_options:
                if col_name in columns:
                    min_value = self.filter_options[col_name].min
                    max_value = self.filter_options[col_name].max
                    include = self.filter_options[col_name].include
                    df = self.filter_df(df,col_name, min_value, max_value, include, columns)
        return df
        

    def filter_df(self,df, col_name,min_value,max_value,include, columns):
        if col_name not in columns:
            return False
        if min_value is not None:
            df = df[df[col_name] >= min_value]
        if max_value is not None:
            df = df[df[col_name] <= max_value]
        if include is not None:
            df = df[df[col_name].isin(include)]
        return df

    def read_hit_table(self, blast_data):
        return pd.read_csv(blast_data,header=None,names=self.blast_columns,sep="\t",low_memory=False)


    def _check_blast_files(self, db_dir: Path, extensions: frozenset):
        """
        """
        logger.info("Verifying blast database: {}".format(str(db_dir)))
        extensions_ = set([i.suffix for i in db_dir.iterdir()])
        if not extensions_.issuperset(extensions):
            raise ValueError("Missing required blast files. {}".format([i for i in extensions_ if i not in extensions]))
    
    def validate_blast_db(self, db_data=None):
        """
        """
        if db_data is None:
            db_data = self.db_data
        if db_data.nucleotide:
            self._check_blast_files(db_data.nucleotide, self.__blast_extensions_nt)

        if db_data.protein:
            self._check_blast_files(db_data.protein, self.__blast_extensions_pt)
    
    def _run_blast(self, db: Path, output: Path):
        """
        db PAth: Path to the blast database to use,
        output Path: Path to file for blast output

        TODO need to use classes db versoin or not pass it too the initializer
        """
        command = [
            self.blast_method,
            '-query', self.query_path,
            '-db', str(db),
            '-out', str(output),
            '-outfmt', "'6 {}'".format(' '.join(self.blast_columns)),
        ]
        for param in self.blast_params:
            if param == "parse_seqids":
                command.append(f"-{param}")
            else:
                command += [f'-{param}', f'{self.blast_params[param]}']
        blast_search_cmd = " ".join([str(x) for x in command])
        logger.info("Blast command: {}".format(blast_search_cmd))
        return run_command(blast_search_cmd)




